/* sysALib.s - system dependent assembly language routines - mz7120 version */

/* Copyright 1988, Wind River Systems, Inc. */

/*
modification history
--------------------
01a,15jan88,miz  written by modifying v01i of hkv2f version.
*/

/*
DESCRIPTION
This module contains system-dependent routines which must be written
in assembly language.  They perform functions such as locking and
unlocking interrupts, trapping to the ROM monitor, etc.

The first routine in this module is the system start-up code.
It is the first code executed after booting.

This module MUST be the first specified on the "ld" command that is used
to build the system.  sysInit is the entry point for UniWorks.

INTERNAL: Many routines in this module "link" and "unlk" the "c" frame pointer
a6@ although they don't use it in any way!  This is only for the benefit of
the stacktrace facility to allow it to properly trace tasks executing within
these routines.

*/

#define ASMLANGUAGE
#include "UniWorks.h"
#include "mz7120.h"
#include "sysLib.h"
#include "config.h"

	/* internals */

	.globl	_sysInit	/* start of system code */
	.globl  _sysKernelTrap
	.globl  _sysVwTrap
	.globl  _sysVwTrapRtn
	.globl	_sysCacheEnable
	.globl	_sysBclrIntA
	.globl	_sysClkIntA
	.globl	_sysAbortIntA

	/* externals */

	.globl	_usrInit	/* system initialization routine */
	.globl	_sysToMonitor	/* exit to monitor */

	.text
	.even

/***********************************************************************
*
* sysInit - start after boot
*
* This is the system start-up entry point for UniWorks in ram.
* It is the first code executed after booting.
* It disables the interrupts, sets up the stack,
* and jumps to the C routine usrInit in usrConfig.c.
*
* The initial stack is set to grow down from sysInit.
* Note that this initial stack is used only by usrInit,
* then is never used again.
* Memory for the stack needs to be accounted for when determining the load
* address of the system.
*
* THIS IS NOT A CALLABLE ROUTINE.

* sysInit ()		/* THIS IS NOT A CALLABLE ROUTINE */

/*
*/

_sysInit:
	movew	#0x3700,sr		/* disable interrupts, turn on M bit */
	movel	#_sysInit,a7		/* set stack to grow down from code */
	movel	#BOOT_WARM_AUTOBOOT,a7@- /* push start type arg = WARM_BOOT */
	jsr	_usrInit		/* never returns - starts up kernel */

/***********************************************************************
*
* sysKernelTrap - trap to kernel function
*
* This routine traps to the kernel. It can only be called from assembly
* language.
* The trap number for the kernel is defined in config.h.
*
* NOMANUAL

* sysKernelTrap ()

*/
_sysKernelTrap:
	trap	#TRAP_KERNEL
	rts

/***********************************************************************
*
* sysVwTrap - trap to UniWorks function
*
* This routine traps to UniWorks. It can only be called from assembly
* language.
* The trap number for UniWorks is defined in config.h.
*
* NOMANUAL

* sysVwTrap ()

*/
_sysVwTrap:
	trap	#TRAP_UNIWORKS
	rts

/*
 * This routine calls the routine whose address is in a0.  Since the system
 * is now in supervisor state (since we got here from a trap) this can be
 * used to call a routine in supervisor state.
 * THIS IS NOT A C CALLABLE ROUTINE!!
 */

_sysVwTrapRtn:
	jsr	a0@		/* vector to proper routine. */
	rte			/* return from trap */

/***********************************************************************
*
* sysCacheEnable - enable cache memory
*
* Enable or disable the 68020 on-chip cache.
*
* SEE ALSO: "MC68020 32-Bit Microprocessor User's Manual"

* VOID sysCacheEnable (enable)
*     BOOL enable;		/* enable if TRUE, otherwise disable */

/*
*/

_sysCacheEnable:

	link	a6,#0
	movel	a6@(0x08),d1		/* put arg in d1 */

	cmpl	#1,d1			/* does arg == TRUE? */
	bne	sce0			/* if not, disable cache */
	movel	#0x09,d0		/* else, clear and enable cache */
	bra	sce1

sce0:	movel	#0,d0			/* disable cache */

	/* Enable or disable the cache by writing to the cache control
	   register.  Since the assemblers differ on the syntax for this one
	   (if they even have it), it's done with two .word's.  Yucko. */

sce1:
	.word	0x4e7b		/* movec ... 			*/
	.word	0x0002		/* .....	d0,cacr 	*/

	unlk	a6
	rts

/************************************************************************
*
* sysBclrIntA - Handle /BCLR interrupt 
*
* This routine releases the VME bus on /BCLR interrupt
*
* NOMANUAL

* VOID sysBclrIntA ()

*/
_sysBclrIntA:
	clrb	0x00180000	/* MZ_BUS_CONTROL_LATCH */
	rte
/************************************************************************
*
* sysClkIntA - Handle heartbeat timer interrupt 
*
* This routine is a stub to catch heartbeat timer interrupts.  It also
* "encourages us to be good VME bus citizens by sharing with the
* other children."  Not a suitable work-around for realtime multiprocessoring.
*
* NOMANUAL

* VOID sysClkIntA ()

*/
_sysClkIntA:
	clrb	0x00180000	/* MZ_BUS_CONTROL_LATCH */
	rte
/************************************************************************
*
* sysAbortIntA - Handle abort switch interrupt 
*
* This routine returns to UniWorks boot monitor on abort interrupt
*
* RETURNS: NONE.

* VOID sysAbortIntA ()

*/
_sysAbortIntA:
	jsr	_sysToMonitor		/* go to monitor land */

