/* if_ln.c - LANCE ethernet network interface for the MVME-147 */

static char *copyright = "Copyright 1988, Wind River Systems, Inc.";

/*
modification history
--------------------
*/

/*
DESCRIPTION
The only user callable routine is:

    lnattach (unit, devAdrs, ivec, ilevel)

*/

#include "UniWorks.h"
#include "iv68k.h"
#include "memLib.h"
#include "ioctl.h"
#include "etherLib.h"

#include "param.h"
#include "mbuf.h"
#include "protosw.h"
#include "socket.h"
#include "errno.h"

#include "if.h"
#include "route.h"

#ifdef	INET
#include "in.h"
#include "in_systm.h"
#include "in_var.h"
#include "ip.h"
#include "if_ether.h"
#endif

#ifdef NS
#include "../netns/ns.h"
#include "../netns/ns_if.h"
#endif

#include "if_ln.h"

IMPORT VOID ipintr ();

IMPORT char lnEnetAddr [];	/* ethernet address to load into lance */

#define MAX_LN		1	/* max number of LANCE chips on MVME-147 */
#define LN_RMD_RLEN	5	/* ring size as a power of 2 */
#define LN_TMD_TLEN	5	/* same for transmit ring */

BOOL ln_logerrors;		/* for logging errors to console */

LOCAL struct ls_softc  *ls_softc [MAX_LN];

/* forward declarations */

LOCAL VOID		lnInt ();
LOCAL VOID		lnIntTask ();

LOCAL VOID		lnInit ();
LOCAL int		lnOutput ();
LOCAL int		lnIoctl ();
LOCAL VOID		lnReset ();

LOCAL struct mbuf      *lnGet ();
LOCAL ln_tmd	       *lnGetFreeTMD ();
LOCAL ln_rmd	       *lnGetFullRMD ();

int tsize = LN_TMD_TLEN;	/* xmit ring size as power of two */
int rsize = LN_RMD_RLEN;	/* recv ring size as power of two */


/*******************************************************************************
*
* lnattach -
*
* Interface exists: make available by filling in network interface
* record.  System will initialize the interface when it is ready
* to accept packets.
*/

STATUS lnattach (unit, devAdrs, ivec, ilevel)
    int	      unit; 
    char     *devAdrs;
    int	      ivec;
    int	      ilevel;

    {
    FAST struct ls_softc  *ls;
    FAST struct ifnet	  *ifp;
    FAST unsigned int	   sz;

    /* allocate basic softc structure */

    ls = (struct ls_softc *)malloc (sizeof (struct ls_softc));
    if (ls == NULL)
	return (ERROR);

    /* zero out entire softc structure */
    bzero ((char *)ls, sizeof (struct ls_softc));

    /* allocate initialization block */

    ls->ib = (ln_ib *)malloc (sizeof (ln_ib));

    if (ls->ib == NULL) 
        goto bad_attach;

    bzero ((char *)ls->ib, sizeof (ln_ib)); /* zero out entire ib */

    if (rsize < 2)		/* limit ring sizes to reasonable values */
        rsize = 2;		/* 4 buffers is reasonable min */

    if (rsize > 7)
        rsize = 7;		/* 128 buffers is max for chip */

    /* allocate receive message descriptor (RMD) ring structure */

    ls->ls_rpo2 = rsize;		/* remember for lnConfig */
    ls->ls_rsize = (1 << rsize);     /* receive message descriptor ring size */

    /* leave room to adjust low three bits */
    sz = ((1 << rsize) + 1) * sizeof (ln_rmd);

    ls->ls_rring = (ln_rmd *)malloc (sz);	/* of pointer to be 000 */
    if (ls->ls_rring == NULL)
        goto bad_attach;

    bzero ((char *)ls->ls_rring, (int)sz);	/* zero out entire RMD ring */

    /* allocate receive buffer space */

    sz = (LN_BUFSIZE << rsize);
    ls->rmd_ring.r_bufs = malloc (sz);	/* room for all the receive buffers */
    if (ls->rmd_ring.r_bufs == NULL)
        goto bad_attach;

    bzero (ls->rmd_ring.r_bufs, (int)sz);	/* zero out entire rbuf area */

    if (tsize < 2)		/* limit ring sizes to reasonable values */
        tsize = 2;			/* 4 buffers is reasonable min */

    if (tsize > 7)
        tsize = 7;			/* 128 buffers is max for chip */

    /* allocate transmit message descriptor (TMD) ring structure */

    ls->ls_tpo2 = tsize;		/* remember for lnConfig */
    ls->ls_tsize = (1 << tsize);   /* transmit message descriptor ring size */

    sz = ((1 << tsize) + 1) * sizeof (ln_tmd);
    ls->ls_tring = (ln_tmd *)malloc (sz);
    if (ls->ls_tring == NULL) 
        goto bad_attach;
    bzero ((char *)ls->ls_tring, (int)sz);	/* zero out entire TMD ring */

    /* allocate transmit buffer space */

    sz = (LN_BUFSIZE << tsize);
    ls->tmd_ring.t_bufs = malloc (sz);	/* room for all the transmit buffers */
    if (ls->tmd_ring.t_bufs == NULL)
        goto bad_attach;
    bzero (ls->tmd_ring.t_bufs, (int)sz);	/* zero out entire tbuf area */

    ls_softc [unit] = ls;		/* remember address for this unit */

    /* initialize device structure */

    ls->ivec    = ivec;
    ls->ilevel  = ilevel;
    ls->devAdrs = (LN_DEVICE *)devAdrs;

    bcopy (lnEnetAddr, (char *)ls->ls_enaddr, sizeof (ls->ls_enaddr));

    ifp = &ls->ls_if;

    ifp->if_unit = unit;
    ifp->if_name = "ln";
    ifp->if_mtu  = 1000;

    lnConfig (ls);			/* initialize LANCE data structures */

    /* attach and enable the LANCE interrupt service routine to vector */

    lnChipReset (ls);				/* reset LANCE */
    (void) intConnect (INUM_TO_IVEC (ivec), lnInt, (int)ls);
    sysLanIntEnable (ilevel);			/* enable LANCE interrupts */
				/* it is not necessarily a VME interrupt */

    ifp->if_init   = lnInit;
    ifp->if_output = lnOutput;
    ifp->if_ioctl  = lnIoctl;
    ifp->if_reset  = lnReset;
    ifp->if_flags  = IFF_BROADCAST;

    if_attach (ifp);

    return (OK);

bad_attach:

    /* free all structures that were allocated above */

    if (ls->tmd_ring.t_bufs != NULL)
        free ((char *)ls->tmd_ring.t_bufs);
    if (ls->ls_tring != NULL)
        free ((char *)ls->ls_tring);

    if (ls->rmd_ring.r_bufs != NULL)
        free ((char *)ls->rmd_ring.r_bufs);
    if (ls->ls_rring != NULL)
        free ((char *)ls->ls_rring);

    if (ls->ib != NULL)
        free ((char *)ls->ib);

    free ((char *)ls);

    return (ERROR);
    }
/*******************************************************************************
*
* lnReset - reset the interface
*
* Mark interface as inactive & reset the chip
*/

LOCAL VOID lnReset (unit)
    int unit;

    {
    FAST struct ls_softc  *ls = ls_softc [unit];

    ls->ls_if.if_flags &= ~IFF_RUNNING;
    lnChipReset (ls);				/* reset LANCE */
    }
/*******************************************************************************
*
* lnInit -
*
* Initialization of interface; clear recorded pending operations.
* Called at boot time (with interrupts disabled?),
* and at ifconfig time via lnIoctl, with interrupts disabled.
*/

LOCAL VOID lnInit (unit)
    int unit;

    {
    FAST struct ls_softc  *ls = ls_softc [unit];
    FAST struct ifnet	  *ifp = &ls->ls_if;

    ifp->if_flags &= ~(IFF_UP | IFF_RUNNING | IFF_PROMISC);

    lnChipReset (ls);				/* disable chip during init */
    lnConfig (ls);				/* reset all ring structures */

    ifp->if_flags |= (IFF_UP | IFF_RUNNING);
    if (ls->promiscuous)
	ifp->if_flags |= (IFF_PROMISC);

    lnChipInit (ls);			/* on return LANCE is running */
    lnStartOutput (ls);			/* tell chip about any pending output */
    }
/*******************************************************************************
*
* lnConfig - fill in initialization block with mode information.
*
* Fill in all fields in the Initialization Block with relevant values.
* In addition, fill in all fields in Transmit Message Descriptor ring
* and Receive Message Descriptor ring.
*/

LOCAL VOID lnConfig (ls)
    FAST struct ls_softc  *ls;

    {
    FAST ln_rmd	  *rmd;
    FAST ln_tmd	  *tmd;
    FAST char	  *buf;
    FAST ln_ib	  *ib;
    int		   i;

    rmd = ls->ls_rring;			/* receive message descriptor ring */
    rmd = (ln_rmd *)(((int)rmd + 7) & ~7);	/* low 3 bits must be 000 */
    ls->ls_rring = rmd;				/* fix softc as well */
    buf = ls->rmd_ring.r_bufs;
    for (i = 0; i < ls->ls_rsize; i++)
        {
	rmd->rbuf_ladr = (u_short)buf;	/* bits 15:00 of buffer address */
	rmd->rbuf_rmd1 = (u_short)((int)buf >> 16) & 0xff;
					/* bits 23:16 of buffer address */

	rmd->rbuf_bcnt = -LN_BUFSIZE;	/* neg of buffer byte count */
	rmd->rbuf_mcnt = 0;		/* no message byte count yet */
	rmd->rbuf_own = 1;		/* buffer now owned by LANCE */
	rmd++;				/* step to next message descriptor */
	buf += LN_BUFSIZE;		/* step to start of next buffer */
	}
    ls->ls_rindex = 0;			/* LANCE will use at start of ring */

    tmd = ls->ls_tring;			/* transmit message descriptor ring */
    tmd = (ln_tmd *)(((int)tmd + 7) & ~7);	/* low 3 bits must be 000 */
    ls->ls_tring = tmd;			/* fix softc as well */
    buf = ls->tmd_ring.t_bufs;
    for (i = 0; i < ls->ls_tsize; i++)
        {
	tmd->tbuf_ladr = (u_short)buf;	/* bits 15:00 of buffer address */
	tmd->tbuf_tmd1 = (u_short)((int)buf >> 16) & 0xff;
					/* bits 23:16 of buffer address */

	tmd->tbuf_bcnt = 0;		/* no message byte count yet */
	tmd->tbuf_tmd3 = 0;		/* no error status yet */
	tmd++;				/* step to next message descriptor */
	buf += LN_BUFSIZE;		/* step to start of next buffer */
	}
    ls->ls_tindex = 0;			/* LANCE will use at start of ring */
    ls->ls_dindex = 0;			/* buffer disposal will lag tindex */

    ib = ls->ib;
    if (ls->promiscuous)
        ib->lnIBMode = 0x8000;	/* chip will be in promiscuous receive mode */
    else
        ib->lnIBMode = 0;	/* chip will be in normal receive mode */

    swab ((char *)ls->ls_enaddr, ib->lnIBPadr, 6);

    ib->lnIBRdraLow = (u_short)ls->ls_rring;
    ib->lnIBRdraHigh = (u_short)(((int)ls->ls_rring >> 16) & 0xff)
      			 | (ls->ls_rpo2 << 13);

    ib->lnIBTdraLow = (u_short)ls->ls_tring;
    ib->lnIBTdraHigh = (u_short)(((int)ls->ls_tring >> 16) & 0xff)
      			 | (ls->ls_tpo2 << 13);
    }
/*******************************************************************************
*
* lnStartOutput - start pending output
*
* Start output to LANCE.
* Queue up all pending datagrams for which transmit buffers are available.
* Kick start the transmitter to avoid the polling interval latency.
* This routine is called by lnInit () and lnOutput ().
* It is very important that this routine be executed with splimp set.
* If this is not done, another task could allocate the same tmd!
*/

LOCAL VOID lnStartOutput (ls)
    FAST struct ls_softc  *ls;

    {
    FAST struct mbuf	  *m;
    struct mbuf		  *mh;
    FAST ln_tmd		  *tmd;
    FAST char		  *buf;
    char		  *buf0;
    struct ether_header	  *eh;
    int			   len;
    int			   s = splimp ();
    int			   oldLevel;

    /* Loop placing message buffers in output ring until no more or no room */

    while (ls->ls_if.if_snd.ifq_head != NULL)
        {
	/* there is something to send */

	if ((tmd = lnGetFreeTMD (ls)) == NULL)	/* get a transmit buffer */
	    break;			/* no transmit buffers available */

        buf = (char *)(tmd->tbuf_ladr | (tmd->tbuf_hadr << 16));
	buf0 = buf;			/* remember where it started */

	/* get message to send */

	IF_DEQUEUE (&ls->ls_if.if_snd, m); /* get head of next mbuf chain */
	mh = m;					/* remember head of chain */

	eh = mtod (m, struct ether_header *);
	bcopy ((char *)ls->ls_enaddr, (char *)eh->ether_shost, 6);

	/* copy packet to write buffer;
	 * note that freeing mbuf is delayed until after we give LANCE 
	 * the buffer so we can overlap operation */

	for (; m != NULL; m = m->m_next)
	    {
	    bcopy (mtod (m, caddr_t), buf, m->m_len);
	    buf += m->m_len;
	    }

	len = max (ETHERMIN + sizeof (struct ether_header), buf - buf0);

	/* call output hook if any */

	if ((etherOutputHookRtn != NULL) &&
	     (* etherOutputHookRtn) (&ls->ls_if, buf0, len))
	    {	
	    /* output hook has already processed this packet */
	    }
	else
	    {
	    /* place a transmit request */

	    oldLevel = intLock ();	/* disable ints during update */

	    tmd->tbuf_tmd3 = 0;		/* clear buffer error status */
	    tmd->tbuf_bcnt = -len;	/* negative message byte count */

	    tmd->tbuf_enp = 1;		/* this buffer is end of packet */
	    tmd->tbuf_stp = 1;		/* this buffer is start of packet */
	    tmd->tbuf_def = 0;		/* clear status bit */
	    tmd->tbuf_one = 0;		/* ditto */
	    tmd->tbuf_more = 0;
	    tmd->tbuf_err = 0;

	    ls->ls_tindex = (ls->ls_tindex + 1) & (ls->ls_tsize - 1);
	    tmd->tbuf_own = 1;		/* give buffer to LANCE */

	    intUnlock (oldLevel);	/* now lnInt won't get confused */

	    lnTransmitDemand (ls);	/* kick start the transmitter */

	    }

	m_freem (mh);		/* now can recycle these */
        }

    splx (s);
    }
/*******************************************************************************
*
* lnInt - handle controller interrupt
*
* This routine is called at interrupt level in response to an interrupt from
* the controller.
*/

LOCAL VOID lnInt (ls)
    FAST struct ls_softc  *ls;

    {
    FAST LN_DEVICE  *dv = ls->devAdrs;
    FAST u_short     stat;
    FAST ln_rmd	    *rmd;
    FAST ln_tmd	    *tmd;

    dv->RAP = 0;			/* select CSR0 */
    stat = dv->lncsr_0 & lncsr_INTMASK;	/* get whole register all at once */
    dv->lncsr_0 = stat;			/* reset all interrupt bits */

    while ((rmd = lnGetFullRMD (ls)) != NULL)
        {				
	/* RMD available */

        netToDoAdd (lnIntTask, ls, rmd);	/* do anything else required */
	ls->ls_rindex = (ls->ls_rindex + 1) & (ls->ls_rsize - 1);
        }

    /* clean up completed transmit commands */

    while (ls->ls_dindex != ls->ls_tindex)
        {
	/* disposal has not caught up */
	tmd = ls->ls_tring + ls->ls_dindex;

	if (tmd->tbuf_own != 0)
	    return;			/* chip owns this one, all done */

	/* buffer not owned by LANCE */

	if (tmd->tbuf_err != 0)
	    ls->ls_if.if_oerrors++;	/* output error */
	else
	    ls->ls_if.if_opackets++;	/* output packet complete */

	tmd->tbuf_tmd1 &= 0xff;		/* clear high byte */
	tmd->tbuf_tmd3 = 0;		/* clear all error & stat stuff */

	/* now bump the tmd disposal index pointer around the ring */
	ls->ls_dindex = (ls->ls_dindex + 1) & (ls->ls_tsize - 1);
        }
    }
/*******************************************************************************
*
* lnIntTask - task level interrupt service
*
* This routine is called at task level indirectly by the interrupt
* service routine to do any message received processing.
*/

LOCAL VOID lnIntTask (ls, rmd)
    struct ls_softc	   *ls;
    FAST ln_rmd		   *rmd;

    {
    lnRecv (ls, rmd);

    rmd->rbuf_rmd1 &= 0xff;		/* clear high byte (errors) */
    rmd->rbuf_mcnt = 0;

    rmd->rbuf_own = 1;			/* give buffer back to LANCE */
    }
/*******************************************************************************
*
* lnRecv -
*
* Process Ethernet receive completion:
*	If input error just drop packet.
*	Otherwise purge input buffered data path and examine 
*	packet to determine type.  If can't determine length
*	from type, then have to drop packet.  Otherwise decapsulate
*	packet based on type and pass to type-specific higher-level
*	input routine.
*/

LOCAL VOID lnRecv (ls, rmd)
    FAST struct ls_softc   *ls;
    FAST ln_rmd		   *rmd;

    {
    FAST struct ether_header  *eh;
    FAST struct mbuf	      *m;
    FAST struct ifqueue	      *inq;
    FAST int		       len, off, resid;
    FAST u_char		      *pData;

    ls->ls_if.if_ipackets++;

    if (rmd->rbuf_err != 0)
	{
    	ls->ls_if.if_ierrors++;
	if (ln_logerrors)
	    logMsg ("ln%d: receive error\n", ls->ls_if.if_unit);
    	return;
	}

    len = rmd->rbuf_mcnt;
    eh = (struct ether_header *)(rmd->rbuf_ladr | (rmd->rbuf_hadr << 16));

    /* call input hook if any */

    if ((etherInputHookRtn != NULL) &&
	(* etherInputHookRtn) (&ls->ls_if, (char *)eh, len))
	{
	return;		/* input hook has already processed this packet */
	}

    /* do software filter if controller is in promiscuous mode */

    if (ls->promiscuous)
	{
	if ((bcmp ((char *)eh->ether_dhost,	/* not our adrs? */
		     (char *)ls->ls_enaddr,
		     sizeof (eh->ether_dhost)) != 0) &&
	    (bcmp ((char *)eh->ether_dhost,	/* not broadcast? */
		     (char *)etherbroadcastaddr,
		     sizeof (eh->ether_dhost)) != 0))
	    return;				/* not for us */
	}
    /*
     * Deal with trailer protocol: if type is trailer
     * get true type from first 16-bit word past data.
     * Remember that type was trailer by setting off.
     */

    if (len >= sizeof (struct ether_header))
        len -= sizeof (struct ether_header);
    else
        len = 0;
    pData = ((u_char *)eh) + (sizeof (struct ether_header));

    eh->ether_type = ntohs ((u_short)eh->ether_type);
    if (eh->ether_type >= ETHERTYPE_TRAIL &&
        eh->ether_type < ETHERTYPE_TRAIL + ETHERTYPE_NTRAILER)
	{
    	off = (eh->ether_type - ETHERTYPE_TRAIL) * 512;

    	if (off >= ETHERMTU)
	    return;				/* sanity */

	eh->ether_type = ntohs (*(u_short *)(pData + off));
	resid          = ntohs (*(u_short *)(pData + off + 2));

    	if (off + resid > len)
	    return;				/* sanity */

    	len = off + resid;
	}
    else
    	off = 0;

    if (len == 0)
        return;					/* sanity */
    /*
     * Pull packet off interface.  Off is nonzero if packet
     * has trailing header; lnGet will then force this header
     * information to be at the front, but we still have to drop
     * the type and length which are at the front of any trailer data.
     */
    m = lnGet ((char *)pData, len, off, (struct ifnet *)&ls->ls_if);
    if (m == 0)
    	return;

    if (off)
	{
    	struct ifnet *ifp;

    	ifp = *(mtod (m, struct ifnet **));
    	m->m_off += 2 * sizeof (u_short);
    	m->m_len -= 2 * sizeof (u_short);
    	*(mtod (m, struct ifnet **)) = ifp;
	}

    switch (eh->ether_type)
	{
#ifdef INET
	case ETHERTYPE_IP:
	    netToDoAdd (ipintr, 0);
	    inq = &ipintrq;
	    break;

	case ETHERTYPE_ARP:
	    arpinput (&ls->ls_ac, m);
	    return;
#endif

#ifdef NS
	case ETHERTYPE_NS:
	    schednetisr (NETISR_NS);
	    inq = &nsintrq;
	    break;
#endif
	default:
	    m_freem (m);
	    return;
	}
    {
    int s = splnet ();

    if (IF_QFULL (inq))
	{
    	IF_DROP (inq);
    	m_freem (m);
    	splx (s);
    	return;
	}

    IF_ENQUEUE (inq, m);

    splx (s);
    }
    }
/*******************************************************************************
* 
* lnGet - Pull read data off an interface.
*
* len is length of data, with local net header stripped.
* off is non-zero if a trailer protocol was used, and
* gives the offset of the trailer information.
* We copy the trailer information and then all the normal
* data into mbufs.  
* Prepend a pointer to the interface structure,
* so that protocols can determine where incoming packets arrived.
*/

LOCAL struct mbuf *lnGet (buf0, totlen, off0, ifp)
    char	 *buf0;
    int		  totlen;
    int		  off0;
    struct ifnet *ifp;

    {
    struct mbuf	       *top;
    struct mbuf	      **mp;
    FAST struct mbuf   *m;
    FAST int		off, len;
    FAST char	       *buf;

    buf = buf0;
    off = off0;

    top = 0;
    mp = &top;

    while (totlen > 0)
	{
	MGET (m, M_DONTWAIT, MT_DATA);
	if (m == 0)
	    {
	    m_freem (top);
	    return (0);
	    }

	if (off)
	    {
	    len = totlen - off;
	    buf += off;
	    }
	else
	    len = totlen;

	m->m_off = MMINOFF;
	if (ifp)
	    {
	    /* Leave room for ifp. */
	    m->m_len = MIN (MLEN - sizeof (ifp), len);
	    m->m_off += sizeof (ifp);
	    }
	else 
	    m->m_len = MIN (MLEN, len);

	bcopy (buf, mtod (m, caddr_t), m->m_len);
	buf += m->m_len;

	*mp = m;
	mp = &m->m_next;

	if (off)
	    {
	    /* sort of an ALGOL-W style for statement... */
	    off += m->m_len;
	    if (off == totlen)
		{
		/* reached end of data; setup to move trailer front */

		buf = buf0;
		off = 0;
		totlen = off0;
		}
	    }
	else
	    totlen -= m->m_len;

	if (ifp)
	    {
	    /* Prepend interface pointer to first mbuf. */
	    m->m_len += sizeof (ifp);
	    m->m_off -= sizeof (ifp);
	    *(mtod (m, struct ifnet **)) = ifp;
	    ifp = (struct ifnet *)0;
	    }
	}

    return (top);
    }
/*******************************************************************************
*
* lnOutput -
*
* Ethernet output routine.
* Encapsulate a packet of type family for the local net.
* Use trailer local net encapsulation if enough data in first
* packet leaves a multiple of 512 bytes of data in remainder.
*/

LOCAL int lnOutput (ifp, m0, dst)
    FAST struct ifnet	*ifp;
    FAST struct mbuf	*m0;
    struct sockaddr	*dst;

    {
    int			   type, s, error;
    u_char		   edst [6];
    struct in_addr	   idst;
    FAST struct ls_softc  *ls = ls_softc [ifp->if_unit];
    FAST struct mbuf	  *m = m0;
    FAST struct ether_header *eh;
    FAST int		   off;
    int			   usetrailers;

    if ((ifp->if_flags & (IFF_UP|IFF_RUNNING)) != (IFF_UP|IFF_RUNNING))
	{
    	error = ENETDOWN;
    	goto bad;
	}

    switch (dst->sa_family)
	{

#ifdef INET
	case AF_INET:
	    idst = ((struct sockaddr_in *)dst)->sin_addr;

	    if (!arpresolve (&ls->ls_ac, m, &idst, edst, &usetrailers))
    		return (0);			/* if not yet resolved */

	    off = ntohs ((u_short)mtod (m, struct ip *)->ip_len) - m->m_len;

	    if (usetrailers && off > 0 && (off & 0x1ff) == 0 &&
		m->m_off >= MMINOFF + 2 * sizeof (u_short))
		{
    		type = ETHERTYPE_TRAIL + (off>>9);
    		m->m_off -= 2 * sizeof (u_short);
    		m->m_len += 2 * sizeof (u_short);
    		*mtod (m, u_short *) = htons (ETHERTYPE_IP);
    		*(mtod (m, u_short *) + 1) = htons ((u_short)m->m_len);
    		goto gottrailertype;
		}

	    type = ETHERTYPE_IP;
	    off = 0;
	    goto gottype;
#endif
#ifdef NS
	case AF_NS:
	    type = ETHERTYPE_NS;
	    bcopy ((caddr_t)&(((struct sockaddr_ns *)dst)->sns_addr.x_host),
    		  (caddr_t)edst, sizeof (edst));
	    off = 0;
	    goto gottype;
#endif

	case AF_UNSPEC:
	    eh = (struct ether_header *)dst->sa_data;
	    bcopy ((caddr_t)eh->ether_dhost, (caddr_t)edst, sizeof (edst));
	    type = eh->ether_type;
	    goto gottype;

	default:
	    printf ("ln%d: can't handle af%d\n", ifp->if_unit, dst->sa_family);
	    error = EAFNOSUPPORT;
	    goto bad;
	}

gottrailertype:
    /*
     * Packet to be sent as trailer: move first packet
     * (control information) to end of chain.
     */
    while (m->m_next)
    	m = m->m_next;
    m->m_next = m0;
    m = m0->m_next;
    m0->m_next = 0;
    m0 = m;

gottype:
    /*
     * Add local net header.  If no space in first mbuf,
     * allocate another.
     */
    if (m->m_off > MMAXOFF || MMINOFF + sizeof (struct ether_header) > m->m_off)
	{
    	m = m_get (M_DONTWAIT, MT_HEADER);
    	if (m == 0)
	    {
	    error = ENOBUFS;
	    goto bad;
	    }
    	m->m_next = m0;
    	m->m_off = MMINOFF;
    	m->m_len = sizeof (struct ether_header);
	}
    else
	{
    	m->m_off -= sizeof (struct ether_header);
    	m->m_len += sizeof (struct ether_header);
	}
    eh = mtod (m, struct ether_header *);
    eh->ether_type = htons ((u_short)type);
    bcopy ((caddr_t)edst, (caddr_t)eh->ether_dhost, sizeof (edst));
    bcopy ((caddr_t)ls->ls_enaddr, (caddr_t)eh->ether_shost, 6);

    /*
     * Queue message on interface, and start output if interface
     * not yet active.
     */

    s = splimp ();
    if (IF_QFULL (&ifp->if_snd))
	{
    	IF_DROP (&ifp->if_snd);
    	splx (s);
    	m_freem (m);
    	return (ENOBUFS);
	}

    IF_ENQUEUE (&ifp->if_snd, m);

    lnStartOutput (ls);

    splx (s);
    return (0);

bad:
    m_freem (m0);
    return (error);
    }
/*******************************************************************************
*
* lnIoctl -
*
* Process an ioctl request.
*/

LOCAL int lnIoctl (ifp, cmd, data)
    FAST struct ifnet	*ifp;
    int			 cmd;
    caddr_t		 data;

    {
    int			   unit = ifp->if_unit;
    FAST struct ls_softc  *ls = ls_softc [unit];
    FAST struct ifaddr	  *ifa = (struct ifaddr *)data;
    int			   s = splimp ();
    int			   error = 0;

    switch (cmd)
	{
	case SIOCSIFADDR:
	    ifp->if_flags |= IFF_UP;
	    lnInit (unit);

	    switch (ifa->ifa_addr.sa_family)
		{
#ifdef	INET
		case AF_INET:
		    ((struct arpcom *)ifp)->ac_ipaddr = IA_SIN (ifa)->sin_addr;
		    arpwhohas ((struct arpcom *)ifp, &IA_SIN (ifa)->sin_addr);
		    break;
#endif	INET
#ifdef	NS
		case AF_NS:
		    {
		    FAST struct ns_addr *ina = &(IA_SNS (ifa)->sns_addr);
    		
		    if (ns_nullhost (*ina))
    			ina->x_host = *(union ns_host *)(ls->ls_enaddr);
		    else
    			lnSetAddr (ina->x_host.c_host,ls);
		    break;
		    }
#endif	NS
		}
	    break;

	case SIOCGIFADDR:
	    bcopy((caddr_t) ls->ls_enaddr,
		  (caddr_t) ((struct ifreq *)data)->ifr_addr.sa_data, 6);
	    break;

	case SIOCGIFFLAGS:
	    *(short *)data = ifp->if_flags;
	    break;

	case SIOCSIFFLAGS:
	    {
	    short flags = *(short *)data;
	    int up      = flags & IFF_UP;
	    int promisc = flags & IFF_PROMISC;

	    if (promisc)
	        ls->promiscuous = TRUE;
	    else
	        ls->promiscuous = FALSE;

	    if ((up != (ifp->if_flags & IFF_UP))
	      | (promisc != (ifp->if_flags & IFF_PROMISC)))
		{
	        if (up)
		    lnInit (unit);		/* going up */
		else
		    lnChipReset (ls);		/* going down */
		}

	    if (ifp->if_flags & IFF_UP)
	        flags |= (IFF_UP|IFF_RUNNING);
	    else
	        flags &= ~(IFF_UP|IFF_RUNNING);

	    ifp->if_flags = flags;
	    }
	    break;

	default:
	    error = EINVAL;
	}

    splx (s);
    return (error);
    }
#ifdef	NS
/*******************************************************************************
*
* lnSetAddr - set ethernet address for unit
*/

LOCAL VOID lnSetAddr (physaddr, ls)
    u_char *physaddr;
    FAST struct ls_softc  *ls;

    {
    int s = splimp ();

    if (physaddr)
    	bcopy ((caddr_t)physaddr, (caddr_t)ls->ls_enaddr, 6);

    if ((ls->ls_if.if_flags & (IFF_UP|IFF_RUNNING)) != 0)
        lnInit (ls->ls_if.if_unit);		/* if running, restart it */

    splx (s);
    }
#endif	NS
/*******************************************************************************
*
* lnChipReset - hardware reset of chip (stop it)
*/

LOCAL VOID lnChipReset (ls)
    struct ls_softc  *ls;

    {
    FAST LN_DEVICE	*dv = ls->devAdrs;

    dv->RAP = 0;				/* select CSR0 */
    dv->lncsr_0 = lncsr_STOP;			/* set the stop bit */
    }
/*******************************************************************************
*
* lnlnChipInit - hardware init of chip (init & start it)
*/

LOCAL VOID lnChipInit (ls)
    FAST struct ls_softc  *ls;

    {
    FAST LN_DEVICE	*dv = ls->devAdrs;

    dv->RAP = 0;		/* select CSR0 */
    dv->lncsr_0 = lncsr_STOP;	/* set the stop bit */
				/* can't write csr1..3 if not stopped */
    dv->RAP = 3;
    dv->lncsr_bswp = 1;		/* addr of word is addr of high byte */
    dv->lncsr_acon = 0;		/* address latch enable, rather than strobe */
    dv->lncsr_bcon = 0;		/* byte mask mode */

    dv->RAP = 2;
    dv->lncsr_iadr_h = (u_short)(((int)ls->ib >> 16) & 0xff);

    dv->RAP = 1;
    dv->lncsr_iadr_l = (u_short)ls->ib;

    dv->RAP = 0;
    dv->lncsr_0 = lncsr_INIT;	/* init chip (read IB) */

    while ((dv->lncsr_0 & lncsr_IDON) == 0)
        ;			/* hang until Initialization DONe */

    ls->ls_rindex = 0;		/* chip will start at beginning */
    ls->ls_tindex = 0;
    ls->ls_dindex = 0;

    dv->lncsr_0 = (lncsr_IDON | lncsr_INEA | lncsr_STRT);
					/* reset IDON state */
					/* enable interrupts from LANCE */
					/* start chip operation */
    }
/*******************************************************************************
*
* lnTransmitDemand - kick start the transmit hardware
*/

LOCAL VOID lnTransmitDemand (ls)
    struct ls_softc  *ls;

    {
    FAST LN_DEVICE	*dv = ls->devAdrs;

    dv->RAP = 0;				/* select CSR0 */
    dv->lncsr_0 = (lncsr_INEA | lncsr_TDMD);	/* kick start the transmitter */
    }
/*******************************************************************************
*
* lnGetFreeTMD - get next available TMD
*/

LOCAL ln_tmd *lnGetFreeTMD (ls)
    FAST struct ls_softc  *ls;

    {
    FAST ln_tmd	  *tmd;
    FAST 	   s;

    /* check if buffer is available (owned by host) */
    /* also check for tindex overrun onto dindex */

    s = splimp ();
    tmd = ls->ls_tring + ls->ls_tindex;
    if ((tmd->tbuf_own != 0)
      | (((ls->ls_tindex + 1) & (ls->ls_tsize - 1)) == ls->ls_dindex))
        tmd = (ln_tmd *) NULL;
    splx (s);

    return (tmd);
    }
/*******************************************************************************
*
* lnGetFullRMD - get next received message RMD
*/

LOCAL ln_rmd *lnGetFullRMD (ls)
    FAST struct ls_softc  *ls;

    {
    FAST ln_rmd	  *rmd;

    /* check if buffer is full (owned by host) */

    rmd = ls->ls_rring + ls->ls_rindex;

    if (rmd->rbuf_own == 0)
        return (rmd);
    else
        return ((ln_rmd *) NULL);
    }
